// 
// 
// 

#include "Weather.h"


void WeatherClass::init()
{
}

void WeatherClass::getWeather() {
	int connectStatus;
	const size_t capacity = 3 * JSON_ARRAY_SIZE(1) + JSON_ARRAY_SIZE(3) + 6 * JSON_OBJECT_SIZE(1) + 4 * JSON_OBJECT_SIZE(2) + 3 * JSON_OBJECT_SIZE(4) + JSON_OBJECT_SIZE(5) + JSON_OBJECT_SIZE(8) + 6 * JSON_OBJECT_SIZE(9) + 450;
	
	getCurrentWeather();
	DynamicJsonDocument doc(capacity);

	DEBUG_PRINTF("Getting Weather forecast\n[Connecting to %s ... ", weatherserver);

	if (connectStatus = client.connect(weatherserver, 80))
	{
		DEBUG_PRINTLN("connected]\n[Sending a request]");
		if (sendRequest(client, weatherserver, FORECAST_3HR) && skipResponseHeaders(client)) {
			String json = client.readStringUntil('\n');
			deserializeJson(doc, json);

			Serial.printf("json 3hr=%s\n", json.c_str()); //DO NOT DEBUG_ AS TOO BIG FOR BUFFER
			//DEBUG_PRINTFS("json 3hr=%s\n", json.c_str());
			int cnt = doc["cnt"]; // 3

			JsonArray list = doc["list"];

			for (int i = 1; i < cnt + 1; i++) {
				JsonObject list_0 = list[i - 1];
				forecast[i].forecastTime = list_0["dt"]; // 1604145600
				forecast[i].forecastTime -= 3600;			//Set to local time

				JsonObject list_0_main = list_0["main"];
				forecast[i].temperature = list_0_main["temp"]; // 23.66
				forecast[i].feelsLike = list_0_main["feels_like"]; // 23.66
				forecast[i].pressure = list_0_main["pressure"]; // 1024
				forecast[i].humidity = list_0_main["humidity"]; // 41
				forecast[i].desiredTemperature = getRequiredTemperature(forecast[i].temperature, forecast[i].forecastTime);
				forecast[i].desiredFeelLikeTemperature = getRequiredTemperature(forecast[i].feelsLike, forecast[i].forecastTime);	//Cater for temperatures to -5 degrees
				forecast[i].dewpoint = dewPoint(forecast[i].temperature, forecast[i].humidity);
				DEBUG_PRINTF(G("FC:%d H:%s, T:%d, P:%d, Hu:%d, Des:%d, Dew:%d\n"), i, ntp.printTime(forecast[i].forecastTime).c_str(), forecast[i].temperature, forecast[i].pressure, forecast[i].humidity, forecast[i].desiredTemperature, forecast[i].dewpoint);
			}
			ss.getTargetTemperature();		//Update target temperature

			DEBUG_PRINTF(G("Setting temperature to %d\n"), ss.targetT);
		}
		DEBUG_PRINTLN("[Disconnected]");
	}
	else
	{
		DEBUG_PRINTF("connection failed! %d, %d\n", connectStatus, client.status());
	}
	client.stop();
	sendWeatherJson();
}

void WeatherClass::getCurrentWeather() {
	const size_t capacity = JSON_ARRAY_SIZE(1) + JSON_OBJECT_SIZE(1) + JSON_OBJECT_SIZE(2) + JSON_OBJECT_SIZE(3) + JSON_OBJECT_SIZE(4) + JSON_OBJECT_SIZE(5) + JSON_OBJECT_SIZE(6) + JSON_OBJECT_SIZE(13) + 270;
	DynamicJsonDocument doc(capacity);
	int connectStatus;
	time_t ft = ntp.getLocalTime_t();

	DEBUG_PRINTLN("Getting current Weather\n");
	if (connectStatus = client.connect(weatherserver, 80))
	{
		DEBUG_PRINTLN("connected]\n[Sending a request]");
		if (sendRequest(client, weatherserver, CURRENT_WEATHER) && skipResponseHeaders(client)) {
			String json = client.readStringUntil('\n');
			deserializeJson(doc, json);

			JsonObject main = doc["main"];
			forecast[0].forecastTime = ft;
			forecast[0].temperature = main["temp"];
			forecast[0].feelsLike = main["feels_like"]; // 23.66
			forecast[0].pressure = main["pressure"];
			forecast[0].humidity = main["humidity"];
			forecast[0].desiredTemperature = getRequiredTemperature(TEMP_EXTERNAL, ft);	//Cater for temperatures to -5 degrees
			forecast[0].desiredFeelLikeTemperature = getRequiredTemperature(forecast[0].feelsLike, ft);	//Cater for temperatures to -5 degrees
			forecast[0].dewpoint = dewPoint(TEMP_EXTERNAL, forecast[0].humidity);

			DEBUG_PRINTF(G("WF: H:%s, T:%d, P:%d, Hu:%d, Des:%d, Dew:%d\n"), ntp.printTime(forecast[0].forecastTime).c_str(), forecast[0].temperature, forecast[0].pressure, forecast[0].humidity, forecast[0].desiredTemperature, forecast[0].dewpoint);
		}
		DEBUG_PRINTLN("[Disconnected]");
	}
	else
	{
		DEBUG_PRINTF("connection failed! %d, %d\n", connectStatus, client.status());
	}
	client.stop();
}

// Send the HTTP GET request to the server
bool WeatherClass::sendRequest(WiFiClient client, const char* host, const char* resource) {
	Serial.print("GET ");
	Serial.println(resource);

	client.print("GET ");
	client.print(resource);
	client.println(" HTTP/1.0");
	client.print("Host: ");
	client.println(host);
	client.println("Connection: close");
	client.println();

	return true;
}

//Calculate projected temperature needed
int WeatherClass::getRequiredTemperature(int temperature, time_t t) {
	int tempCalc;

	tempCalc = heatingRequirements[temperature + 10]; //+10 is for array offset from 0 (so temperature scale -10 -> ...
	DEBUG_PRINTF(G("getRequiredTemperature: in:%d, tempCalc:%d\n"), temperature, tempCalc);
	//Reduce temperature overnight
	if (!ntp.isTimeDaytime(t))
		tempCalc -= ss.nightTemperature;
	DEBUG_PRINTF(G("getRequiredTemperature(2): in:%d, tempCalc:%d\n"), temperature, tempCalc);
	return tempCalc;
}

//Send weather information to browser
void WeatherClass::sendWeatherJson() {
	const size_t capacity = JSON_OBJECT_SIZE(22) + 340;
	DynamicJsonDocument root(capacity);
	char temp[10];
	String jsonString;

	root["lastWeather"] = ntp.printTime(forecast[0].forecastTime);
	snprintf(temp, sizeof(temp), "%d (%s)", forecast[0].desiredTemperature, ntp.isTimeDaytime(forecast[0].forecastTime) ? "D" : "N");
	root["currentD"] = temp;
	snprintf(temp, sizeof(temp), "A:%d F:%d", TEMP_EXTERNAL, forecast[0].temperature);
	root["currentF"] = temp;
	root["currentRh"] = forecast[0].humidity;
	root["currentDew"] = forecast[0].dewpoint;
	root["currentFeelsLike"] = forecast[0].feelsLike;

	root["time1"] = ntp.printTime(forecast[1].forecastTime);
	root["temp1F"] = forecast[1].temperature;
	snprintf(temp, sizeof(temp), "%d (%s)", forecast[1].desiredTemperature, ntp.isTimeDaytime(forecast[1].forecastTime) ? "D" : "N");
	root["temp1D"] = temp;
	root["rh1"] = forecast[1].humidity;
	root["dewpoint1"] = forecast[1].dewpoint;
	root["feelsLike1"] = forecast[1].feelsLike;

	root["time2"] = ntp.printTime(forecast[2].forecastTime);
	root["temp2F"] = forecast[2].temperature;
	snprintf(temp, sizeof(temp), "%d (%s)", forecast[2].desiredTemperature, ntp.isTimeDaytime(forecast[2].forecastTime) ? "D" : "N");
	root["temp2D"] = temp;
	root["rh2"] = forecast[2].humidity;
	root["dewpoint2"] = forecast[2].dewpoint;
	root["feelsLike2"] = forecast[2].feelsLike;

	root["time3"] = ntp.printTime(forecast[3].forecastTime);
	root["temp3F"] = forecast[3].temperature;
	snprintf(temp, sizeof(temp), "%d (%s)", forecast[3].desiredTemperature, ntp.isTimeDaytime(forecast[3].forecastTime) ? "D" : "N");
	root["temp3D"] = temp;
	root["rh3"] = forecast[3].humidity;
	root["dewpoint3"] = forecast[3].dewpoint;
	root["feelsLike3"] = forecast[3].feelsLike;

	serializeJson(root, jsonString);
	ws.textAll(jsonString);
}

// dewPoint function NOAA
// reference (1) : http://wahiduddin.net/calc/density_algorithms.htm
// reference (2) : http://www.colorado.edu/geography/weather_station/Geog_site/about.htm
//
double WeatherClass::dewPoint(double celsius, double humidity)
{
	//DEBUG_PRINTF(G("Dewpoint in:%d, %d"), celsius, humidity);
	// (1) Saturation Vapor Pressure = ESGG(T)
	double RATIO = 373.15 / (273.15 + celsius);
	double RHS = -7.90298 * (RATIO - 1);
	RHS += 5.02808 * log10(RATIO);
	RHS += -1.3816e-7 * (pow(10, (11.344 * (1 - 1 / RATIO))) - 1);
	RHS += 8.1328e-3 * (pow(10, (-3.49149 * (RATIO - 1))) - 1);
	RHS += log10(1013.246);

	// factor -3 is to adjust units - Vapor Pressure SVP * humidity
	double VP = pow(10, RHS - 3) * humidity;

	// (2) DEWPOINT = F(Vapor Pressure)
	double T = log(VP / 0.61078);   // temp var
	double res = (241.88 * T) / (17.558 - T);
	return truncate((float)res, 2);
}

float WeatherClass::truncate(float val, byte dec)
{
	float x = val * pow(10, dec);
	float y = round(x);
	float z = x - y;
	if ((int)z == 5)
	{
		y++;
	}
	else {}
	x = y / pow(10, dec);
	return x;
}

// Skip HTTP headers so that we are at the beginning of the response's body
bool WeatherClass::skipResponseHeaders(WiFiClient client) {
	// HTTP headers end with an empty line
	char endOfHeaders[] = "\r\n\r\n";

	client.setTimeout(HTTP_TIMEOUT);
	bool ok = client.find(endOfHeaders);

	if (!ok) {
		Serial.println("No response or invalid response!");
	}

	return ok;
}

WeatherClass Weather;

