// 
// 
// 

#include "hpWebServer.h"
#include <ESP8266SSDP.h>
#include <StreamString.h>

//SSDP properties
constexpr auto modelName = "Heatpump";
constexpr auto modelNumber = "Diiatsu";

AsyncWebServer server(80);

AsyncWebSocket ws("/ws");
AsyncWebSocket wsd("/wsd");

void sendStatusPage(AsyncWebServerRequest* request) {
	request->send(FileFS, "/status.html");
	//ss.refreshData = true;
	//ss.bMain = true;
}

/////////////////////////////////////  SSDP network discovery //////////////////////////////////////
void setupSSDP() {
	SSDP.setSchemaURL("description.xml");
	SSDP.setHTTPPort(80);
	SSDP.setDeviceType("upnp:rootdevice");
	SSDP.setModelName(modelName);
	SSDP.setModelNumber(modelNumber);
	SSDP.begin();

	server.on("/description.xml", HTTP_GET, [](AsyncWebServerRequest* request) {
		char hostName[16] = "Heatpump";
		if (strlen(ss.hostName) > 0)
			snprintf(hostName, sizeof(hostName), "Heatpump-%s", ss.hostName);

		StreamString output;
		if (output.reserve(1024)) {
			IPAddress ip = WiFi.localIP();
			uint32_t chipId = ESP.getChipId();
			output.printf(ssdpTemplate,
				ip[0], ip[1], ip[2], ip[3],
				hostName,
				ip[0], ip[1], ip[2], ip[3],
				chipId,
				modelName,
				modelNumber,
				(uint8_t)((chipId >> 16) & 0xff),
				(uint8_t)((chipId >> 8) & 0xff),
				(uint8_t)chipId & 0xff
			);
			request->send(200, "text/xml", (String)output);
		}
		else {
			request->send(500);
		}
		});
}

void refreshPage(AsyncWebServerRequest* request) {
	DEBUG_PRINTLN(G("Refresh page data"));
	ss.refreshData = true;		//Force a data update
	ss.bMain = true;
	request->send(200);
}

void onRequest(AsyncWebServerRequest* request) {
	//Handle Unknown Request
	request->send(404);
}

void onBody(AsyncWebServerRequest* request, uint8_t* data, size_t len, size_t index, size_t total) {
	//Handle body
}

void onUpload(AsyncWebServerRequest* request, String filename, size_t index, uint8_t* data, size_t len, bool final) {
	//Handle upload
}

void onEvent(AsyncWebSocket* server, AsyncWebSocketClient* client, AwsEventType type, void* arg, uint8_t* data, size_t len) {
	//Handle WebSocket event
}

void setupWebServer()
{
	// Start and init debug winsocket
	ws.onEvent([](AsyncWebSocket* server, AsyncWebSocketClient* client, AwsEventType type, void* arg, uint8_t* data, size_t len)
		{
			//Serial.printf("wsd event: %d\n", type);
			if (type == WS_EVT_CONNECT)
			{
				//Serial.printf("wsd[%s][%u] connected\n", server->url(), client->id());
				client->ping();
			}
		});
	server.addHandler(&ws);

	// Start and init debug winsocket
	wsd.onEvent([](AsyncWebSocket* server, AsyncWebSocketClient* client, AwsEventType type, void* arg, uint8_t* data, size_t len)
		{
			//Serial.printf("wsd event: %d\n", type);
			if (type == WS_EVT_CONNECT)
			{
				//Serial.printf("wsd[%s][%u] connected\n", server->url(), client->id());
				client->ping();
			}
		});
	server.addHandler(&wsd);

	server.on("/", HTTP_GET, [](AsyncWebServerRequest* request) {sendStatusPage(request); });
	server.on("/status", HTTP_GET, [](AsyncWebServerRequest* request) {sendStatusPage(request); });
	server.on("/refresh", HTTP_GET, [](AsyncWebServerRequest* request) {refreshPage(request); });
	
	server.on("/getset", HTTP_GET, [](AsyncWebServerRequest* request) {ss.readSettings(); request->send(200); });
	server.on("/values", HTTP_GET, [](AsyncWebServerRequest* request) {ss.debugOutParameters(); request->send(200); });
	server.on("/getWeather", HTTP_GET, [](AsyncWebServerRequest* request) {
		request->send(200);
		ss.bGetWeather = true;
		});

	server.on("/debug", HTTP_GET, [](AsyncWebServerRequest* request) {request->send(FileFS, "/debug.html"); });
	server.on("/fon", HTTP_GET, [](AsyncWebServerRequest* request) {ss.setRelay(FAN, ON);  request->send(200); });
	server.on("/foff", HTTP_GET, [](AsyncWebServerRequest* request) {ss.setRelay(FAN, OFF);  request->send(200); });
	server.on("/von", HTTP_GET, [](AsyncWebServerRequest* request) {ss.setRelay(VALVE, ON);  request->send(200); });
	server.on("/voff", HTTP_GET, [](AsyncWebServerRequest* request) {ss.setRelay(VALVE, OFF);  request->send(200); });
	server.on("/con", HTTP_GET, [](AsyncWebServerRequest* request) {ss.setRelay(COMPRESSOR, ON);  request->send(200); });
	server.on("/coff", HTTP_GET, [](AsyncWebServerRequest* request) {ss.setRelay(COMPRESSOR, OFF);  request->send(200); });
	server.on("/pon", HTTP_GET, [](AsyncWebServerRequest* request) {ss.setRelay(PUMP, ON);  request->send(200); });
	server.on("/poff", HTTP_GET, [](AsyncWebServerRequest* request) {ss.setRelay(PUMP, OFF);  request->send(200); });

	server.on("/setup", HTTP_GET, [](AsyncWebServerRequest* request) {
		//if (strcmp(ss.www_username, "admin") + strcmp(ss.www_password, "admin") != 0) {
		//	if (!request->authenticate(ss.www_username, ss.www_password)) {
		//		DEBUG_PRINTLN(G("Not authenticated"));
		//		return request->requestAuthentication();
		//	}
		//}
		request->send(FileFS, "/setup.html"); });

	// Send username and password
	server.on("/shuspw", HTTP_GET, [](AsyncWebServerRequest* request) {request->send(200);
	DEBUG_PRINTF("U:%s P:%s\n", ss.www_username, ss.www_password); });

	server.on("/settings", HTTP_GET, [](AsyncWebServerRequest* request) {
		DEBUG_PRINTLN(G("request settings"));
		request->send(200, "text/xml", ss.buildSettingsJson());
		ss.refreshData = true;
		});

	server.on("/saveSettings", HTTP_POST, [](AsyncWebServerRequest* request) {
		DEBUG_PRINTLN(G("saveSettings"));

		if (request->args() > 0)
		{
			ss.JsonToSave = request->arg("z");
			ss.bSaveSettings = true;
			request->send(200, F("text/plain"), F("Saved settings"));
		}
		});

	server.on("/devReboot", HTTP_GET, [](AsyncWebServerRequest* request) {
		request->send(200, "text/plain", "Rebooting...");
		ss.shouldReboot = true;
		});

	server.on("/mac", HTTP_GET, [](AsyncWebServerRequest* request) {
		request->send(200, "");
		DEBUG_PRINTF(G("MAC:%s\n"), WiFi.macAddress().c_str());
		});

	server.on("/hpstate", HTTP_GET, [](AsyncWebServerRequest* request) {
		request->send(200, "text/xml", ss.getSystemStatus(true));
		});

	server.on("/pwr", HTTP_POST, [](AsyncWebServerRequest* request) {
		if (request->args() > 0)
		{
			bool newMode = ss.masterOnOff;
			ss.masterOnOff = (int)request->arg("p").toInt();
			if (newMode != ss.masterOnOff)
			{
				ss.powerOnOff();
			}
		}
		request->send(200);
		});

	//Change operating mode - set a flag so correct sequence can be performed
	server.on("/mode", HTTP_POST, [](AsyncWebServerRequest* request) {
		if (request->args() > 0)
		{
			bool newMode = ss.operatingMode;
			ss.operatingMode = (int)request->arg("m").toInt();
			if (newMode != ss.operatingMode)
			{
				DEBUG_PRINTF("mode %s\n", oMode[ss.operatingMode]);
				ss.millisNow = millis();
				ss.startupDelay = true;
				eepromUtil.eeprom_write_byte(E_OPERATINGMODE, ss.operatingMode);
				EEPROM.commit();
			}
		}
		ss.getTargetTemperature();		//Update target temperature
		request->send(200);
		});

	server.on("/logout", HTTP_GET, [](AsyncWebServerRequest* request) {request->requestAuthentication(); });
	// Reset all system parameters to default
	server.on("/resetParam", HTTP_GET, [](AsyncWebServerRequest* request) {
		ss.resetParameters();
		request->send(200, "text/plain", "All parameters reset to default");
		});


	/////////////////////////////////////////////  Network   ////////////////////////////////////////////////////
	server.on("/netset", HTTP_GET, [](AsyncWebServerRequest* request) {
		DEBUG_PRINTLN(G("network setup"));
		request->send(FileFS, "/netset.html");
		});

	server.on("/getNet", HTTP_GET, [](AsyncWebServerRequest* request) {
		DEBUG_PRINTLN(G("read network settings"));
		String strJson = ss.buildNetSettingsJson();
		request->send(200, "text/xml", strJson);
		});

	server.on("/saveNet", HTTP_POST, [](AsyncWebServerRequest* request) {
		if (request->args() > 0)
		{
			ss.JsonToSave = request->arg("a");
			ss.bSaveNetSettings = true;
		}
		request->send(200);
		});

	//Serve all files. 
	server.serveStatic("/", FileFS, "/", "max-age=2592000");

	// Catch-All Handlers
	// Any request that can not find a Handler that canHandle it ends in the callbacks below.
	server.onNotFound(onRequest);
	server.onFileUpload(onUpload);
	server.onRequestBody(onBody);
	setupSSDP();
	server.begin();
}

void setupOTAUpdate(void)
{
	// No authentication by default
	ArduinoOTA.setPassword(OTA_PASSWORD);

	ArduinoOTA.onStart([]() {
		DEBUG_PRINTLN(G("Start update"));
		});
	ArduinoOTA.onEnd([]() {
		DEBUG_PRINTLN(G("\nEnd"));
		});
	ArduinoOTA.onProgress([](unsigned int progress, unsigned int total) {
		char updateProgress[50];
		unsigned int percent = progress / (total / 100);
		snprintf(updateProgress, sizeof(updateProgress), G("Progress: %u%%"), (progress / (total / 100)));
		if (percent % 10 == 0) {		//Only every 10%
			DEBUG_PRINTF("%s\n", updateProgress);
		}
		ss.last_loop = millis();
		});
	ArduinoOTA.onError([](ota_error_t error) {
		DEBUG_PRINTF("Error[%u]: ", error);
		if (error == OTA_AUTH_ERROR) DEBUG_PRINTLN(G("Auth Failed"));
		else if (error == OTA_BEGIN_ERROR) DEBUG_PRINTLN(G("Begin Failed"));
		else if (error == OTA_CONNECT_ERROR) DEBUG_PRINTLN(G("Connect Failed"));
		else if (error == OTA_RECEIVE_ERROR) DEBUG_PRINTLN(G("Receive Failed"));
		else if (error == OTA_END_ERROR) DEBUG_PRINTLN(G("End Failed"));
		});
	ArduinoOTA.begin();
}

